package de.hdm_stuttgart.mi.bb;

import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;
import org.json.JSONObject;

import javax.ws.rs.core.Response;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;

public class Id2Topic {

    static final private Logger log = LogManager.getLogger(Id2Topic.class);

    static {
        // Disable startup info
        final java.util.logging.Logger mongoLogger = java.util.logging.Logger.getLogger( "org.mongodb.driver" );
        mongoLogger.setLevel(java.util.logging.Level.SEVERE); // e.g. or Log.WARNING, etc.

        final String
                configPathName= System.getProperty("user.home"),
                configFileBaseName = "bbtopics",
                configFileFullyQualifiedName = configPathName + '/' + configFileBaseName + ".properties";
        log.info("Reading config file '" + configFileFullyQualifiedName + "'");
        final File file = new File(configPathName);

        ResourceBundle config;
        try {
            final URL[] urls = new URL[]{file.toURI().toURL()};
            final ClassLoader loader = new URLClassLoader(urls);
            config = ResourceBundle.getBundle(configFileBaseName, Locale.getDefault(), loader);
        } catch (final MalformedURLException | MissingResourceException e) {
            config = null;
            log.fatal("Unable to open URI " + configFileFullyQualifiedName, e);
            System.exit(1);
        }
        nodebbAdminUid = config.getString("nodebbAdminUid");
        bearerToken = config.getString("nodebbBearerToken");
        apiUrl = config.getString("nodebbWriteApiTopicsEndpoint");
        mongoNodebbDatabase = config.getString("mongoNodebbDatabase");

        updateTopicTitles = Boolean.parseBoolean(config.getString("updateTopicTitles").trim());

        db = new ModifyNodebbTopic(
                config.getString("mongoUri"),
                config.getString("mongoNodebbDatabase"));
    }
    static final String apiUrl, mongoNodebbDatabase, nodebbAdminUid, bearerToken;

    static final boolean updateTopicTitles;

    static private final ModifyNodebbTopic db;

    static private final String
            KEY_PAYLOAD = "payload",
            KEY_TOPICDATA = "topicData";

    public static void main(String[] args) {

        log.info(updateTopicTitles ? "Updating existing topic's titles": "Not updating existing topic's titles");

        final SAXBuilder saxBuilder = new SAXBuilder();

        if (0 == args.length) {
            System.err.println("Usage: id2topic <docbook2topic basefilename>");
        } else {
            final File inputFile = new File(args[0] + ".xml");

            try {
                final Document document = saxBuilder.build(inputFile);
                handleDocbok2idMap(document.getRootElement());

                final Format outFormat = Format.getPrettyFormat();

                // Serialize to console
                final XMLOutputter printer = new XMLOutputter(outFormat);

                final String outputFileName = args[0] + ".tid.xml";
                final FileWriter outputFile = new FileWriter(outputFileName);
                printer.output(document.getRootElement(), outputFile);
            } catch (JDOMException e) {
                System.err.println("Unable to parse " + inputFile);
            } catch (IOException e) {
                System.err.println("Unable to read " + inputFile);
            }
        }
    }

    static private void handleDocbok2idMap(final Element map) {
        final String mapPrefix = map.getAttributeValue("prefix");

        for (final Element entry : map.getChildren("entry")) {

            final int tid = createTopic(
                    mapPrefix + ":" + entry.getAttributeValue("id"),
                    entry.getAttributeValue("forumId"),
                    entry.getChildText("title"),
                    entry.getChildText("content"));

            entry.setAttribute("tid", Integer.toString(tid));
        }
    }

    static private int createTopic(final String hdmId, final String categoryId, final String title, final String content) {

        final int tid = db.hdmIdExists(hdmId);
        if (tid < 0) { // No related topic yet

            final CloseableHttpClient client = HttpClientBuilder.create().build();

            final HttpPost post = new HttpPost(apiUrl);
            post.setHeader("Accept-Charset","UTF-8");
            post.setHeader("Authorization", "Bearer " + bearerToken);

            try {
                final List<NameValuePair> topicInstance = new ArrayList<>(4);
                topicInstance.add(new BasicNameValuePair("_uid", nodebbAdminUid));
                topicInstance.add(new BasicNameValuePair("cid", categoryId));
                topicInstance.add(new BasicNameValuePair("title", title));
                topicInstance.add(new BasicNameValuePair("content", content));
                post.setEntity(new UrlEncodedFormEntity (topicInstance, "UTF-8"));

                final HttpResponse response = client.execute(post);
                final JSONObject newTopic = new JSONObject(EntityUtils.toString(response.getEntity()));

                if (Response.Status.OK.getStatusCode() != response.getStatusLine().getStatusCode()) {
                    log.fatal(response.getStatusLine());
                    log.fatal("id=" + hdmId + ":" + newTopic.toString() + ", exiting!");
                    System.exit(1);
                }
                int newTid  = (int) newTopic.getJSONObject(KEY_PAYLOAD).getJSONObject(KEY_TOPICDATA).get("tid");
                log.info("Created topic " + newTid + " (hdmId = " + hdmId + ", tid = " + newTid + ")");
                db.addHdmId(newTid, hdmId);
                return newTid;
            } catch (IOException e) {
                e.printStackTrace();
            }
            return -1;
        } else {
            log.info ("Found topic (hdmId = " +  hdmId + ", tid = " + tid + ")");
            if (updateTopicTitles) {
                db.updateTopicTitle(tid, title, content);
            }
            return tid;
            // Possibly aligning title and content with lecture notes?
            // Difficult: Incomplete NodeBB write API --> required MongoDB fiddling

            /*
            final WebTarget hdmTtopicTarget =
                    client.target(config.getString("nodebbReadApiTopicsEndpoint") + "/" + tid);

            final Response response = hdmTtopicTarget.request(MediaType.APPLICATION_JSON_TYPE).get();

            final Map<String, Object> hdmTopicMap = response.readEntity(Map.class);
            log.info("-----------------------Topic [tid=" + tid + "] data:");
            for (String k : hdmTopicMap) {
                System.out.println("key '" + k + "', val '" + m.get(k) + "'");
            }
             */
        }
    }
}